import java.util.concurrent.*;
import java.util.concurrent.locks.*;
import java.util.Iterator;

abstract class Rom {
    final int ROM_NR, ROM_KVM, ETASJE, ANT_SENGEPLASSER;
    boolean erLedig = true;
    Rom nesteRom = null;
    Rom nesteRomEtasje = null;

    Rom (int nr, int kvm, int etasje, int antSeng) {
	ROM_NR = nr;  ROM_KVM = kvm;
	ETASJE = etasje;  ANT_SENGEPLASSER = antSeng;
    }

    // Hjelpemetode for å finne egnet rom:
    boolean erEgnet (int antSeng, boolean kjokken) {
	if (! erLedig) return false;
	if (antSeng != ANT_SENGEPLASSER) return false;
	if (kjokken) return false;
	return true;
    }

    @Override
    public String toString () {
	return ROM_NR + " (" + ETASJE + " etasje) med " + ROM_KVM + " m2 og " +
	    ANT_SENGEPLASSER + " sengeplasser [" +
	    (erLedig ? "ledig" : "opptatt") + "]";
    }
}

class EnkeltRom extends Rom {
    EnkeltRom (int nr, int kvm, int etasje, int antSeng) {
	super(nr, kvm, etasje, antSeng);
    }

    @Override
    public String toString () {
	return "Enkeltrom " + super.toString();
    }
}

class VanligRom extends Rom {
    VanligRom (int nr, int kvm, int etasje, int antSeng) {
	super(nr, kvm, etasje, antSeng);
    }

    @Override
    public String toString () {
	return "Vanlig rom " + super.toString();
    }
}

class Suite extends Rom {
    Suite (int nr, int kvm, int etasje, int antSeng) {
	super(nr, kvm, etasje, antSeng);
    }

    @Override
    public String toString () {
	return "Suite " + super.toString();
    }
}

interface Kjokken {
    int hentKjokkenStorrelse ();
}

class VanligRomMedKjokken extends VanligRom implements Kjokken {
    final int KJOKKEN_KVM;

    VanligRomMedKjokken (int nr, int kvm, int etasje, int antSeng, int kjKvm) {
	super(nr, kvm, etasje, antSeng);
	KJOKKEN_KVM = kjKvm;
    }

    @Override
    public int hentKjokkenStorrelse () {
	return KJOKKEN_KVM;
    }

    // Hjelpemetode for å finne egnet rom:
    @Override
    boolean erEgnet (int antSeng, boolean kjokken) {
	if (! erLedig) return false;
	if (antSeng != ANT_SENGEPLASSER) return false;
	return true;
    }

    @Override
    public String toString() {
	return super.toString() + " med " + KJOKKEN_KVM + " m2 kjokken";
    }

}

class SuiteMedKjokken extends Suite implements Kjokken {
    final int KJOKKEN_KVM;

    SuiteMedKjokken (int nr, int kvm, int etasje, int antSeng, int kjKvm) {
	super(nr, kvm, etasje, antSeng);
	KJOKKEN_KVM = kjKvm;
    }

    @Override
    public int hentKjokkenStorrelse () {
	return KJOKKEN_KVM;
    }

    // Hjelpemetode for å finne egnet rom:
    @Override
    boolean erEgnet (int antSeng, boolean kjokken) {
	if (! erLedig) return false;
	if (antSeng != ANT_SENGEPLASSER) return false;
	return true;
    }

    @Override
    public String toString() {
	return super.toString() + " med " + KJOKKEN_KVM + " m2 kjokken";
    }
}

class Hotell implements Iterable<Rom> {
    final int MAX_ANT_SENGEPLASSER = 8;
    final int ANTALL_ETASJER;
    Rom forsteRom = null;
    Reservasjon forsteR = null, sisteR = null;
    Rom[] forsteRomEtasje;

    Hotell (int antEt) {
	ANTALL_ETASJER = antEt;
	forsteRomEtasje = new Rom[ANTALL_ETASJER+1];
    }

    // Hjelpemetode: Finn en gjests reservasjon:
    Reservasjon finnRes (String navn) {
	Reservasjon p = forsteR;
	while (p != null) {
	    if (p.gjest.navn.equals(navn))
		return p;
	    p = p.nesteR;
	}
	return null;
    }

    void tildelRom (String navn) {
	// Finn reservasjonen:
	Reservasjon res = finnRes(navn);
	if (res == null) throw new IngenReservasjon(navn);
	taUtRes(res);

	// Finn egnet rom:
	Rom rom = null;
	for (int sx = res.trengerSenger;  sx <= MAX_ANT_SENGEPLASSER;  ++sx) {
	    rom = finnRom(sx, res.onskerKjokken);
	    if (rom != null) break;
	}
	if (rom == null)
	    throw new IntetLedigRom(navn, res.trengerSenger);

	// Tildel rommet:
	rom.erLedig = false;
	res.gjest.rom = rom;
	System.out.println(navn + " har fatt tildelt " + rom);
    }

    Rom finnRom (int antSeng, boolean kjokken) {
	Rom romP = forsteRom;
	while (romP != null) {
	    if (romP.erEgnet(antSeng,kjokken))
		return romP;
	    romP = romP.nesteRom;
	}
	return null;
    }

    void taUtRes (Reservasjon r) {
	if (r == forsteR && forsteR == sisteR) {
	    forsteR = sisteR = null; 
	} else if (r == forsteR) {
	    forsteR = forsteR.nesteR;
	    forsteR.forrigeR = null;
	} else if (r == sisteR) {
	    sisteR = sisteR.forrigeR;
	    sisteR.nesteR = null;
	} else {
	    r.forrigeR.nesteR = r.nesteR;
	    r.nesteR.forrigeR = r.forrigeR;
	}
	r.forrigeR = r.nesteR = null;
    }

    class RomIterator implements Iterator<Rom> {
	int denneEtasjen;
	Rom detteRommet;

	RomIterator () {
	    denneEtasjen = 0;
	    detteRommet = forsteRomEtasje[0];
	    finnEtRom();
	}

	void finnEtRom () {
	    while (denneEtasjen<=ANTALL_ETASJER && detteRommet==null) {
		++denneEtasjen;
		if (denneEtasjen <= ANTALL_ETASJER)
		    detteRommet = forsteRomEtasje[denneEtasjen];
	    }
	}

	@Override
	public boolean hasNext () {
	    return denneEtasjen <= ANTALL_ETASJER;
	}

	@Override
	public Rom next () {
	    Rom svar = detteRommet;
	    detteRommet = detteRommet.nesteRomEtasje;
	    finnEtRom();
	    return svar;
	}
    }

    public RomIterator iterator () {
	return new RomIterator();
    }

    int[] ledigeRom () {
	int[] ledige = new int[MAX_ANT_SENGEPLASSER];
	for (Rom r: this)
	    if (r.erLedig) ++ledige[r.ANT_SENGEPLASSER-1];
	return ledige;
    }

    // Testprogram (ikke bedt om i oppgaven):
    public static void main (String[] arg) {
	Hotell h = new Hotell(4);
	Rom rom1 = new EnkeltRom(1, 15, 0, 1);
	Rom rom2 = new VanligRom(21, 20, 2, 2);
	Rom rom3 = new VanligRomMedKjokken(22, 25, 2, 3, 6);
	Rom rom4 = new SuiteMedKjokken(31, 40, 3, 6, 10);
	h.forsteRom = rom1;  rom1.nesteRom = rom2;
	rom2.nesteRom = rom3; rom3.nesteRom = rom4;
	h.forsteRomEtasje[0] = rom1;
	h.forsteRomEtasje[2] = rom2;  rom2.nesteRomEtasje = rom3;
	h.forsteRomEtasje[3] = rom4;

	System.out.println("Rommene er;");
	for (Rom r: h)
	    System.out.println("  " + r);
	System.out.println();

	Gjest g1 = new Gjest();  g1.navn = "Dag";
	Gjest g2 = new Gjest();  g2.navn = "Siri";
	Gjest g3 = new Gjest();  g3.navn = "Stein";
	Reservasjon res1 = new Reservasjon();
	res1.gjest = g1;  res1.trengerSenger = 5;  res1.onskerKjokken = false;
	Reservasjon res2 = new Reservasjon();
	res2.gjest = g2;  res2.trengerSenger = 2;  res2.onskerKjokken = true;
	Reservasjon res3 = new Reservasjon();
	res3.gjest = g3;  res3.trengerSenger = 2;  res3.onskerKjokken = true;
	h.forsteR = res1;  h.sisteR = res3;
	res1.nesteR = res2;  res2.nesteR = res3;
	res2.forrigeR = res1;  res3.forrigeR = res2;

	Hotellkjede kjede = new Hotellkjede();
	kjede.alleHoteller[0] = h;
	kjede.alleHoteller[1] = new Hotell(2);
	kjede.skrivUtLedigeRomMedTrader();
	System.out.println();

	String[] navneliste = { "Dag", "Anne", "Stein", "Siri" };
	for (String s: navneliste) {
	    try {
		h.tildelRom(s);
	    } catch (IngenReservasjon e) {
		System.out.println(e);
	    } catch (IntetLedigRom e) {
		System.out.println(e);
	    }
	}
	System.out.println();
	kjede.skrivUtLedigeRomMedTrader();
    }
}

class Hotellkjede {
    final int ANTALL_HOTELLER = 2;
    final int KJEDE_MAKS_ANTALL_SENGEPLASSER = 10;
    Hotell[] alleHoteller = new Hotell[ANTALL_HOTELLER];

    void skrivUtLedigeRomMedTrader () {
	Monitor mon = new Monitor(KJEDE_MAKS_ANTALL_SENGEPLASSER);
	CountDownLatch teller = new CountDownLatch(ANTALL_HOTELLER);
	for (Hotell h: alleHoteller)
	    new Thread(new RomTeller(h,mon,teller)).start();
	try {
	    teller.await();
	} catch (InterruptedException e) {}
	
	System.out.println("Totalt antall ledige rom:");
	int[] ledige = mon.hentLedigeRom();
	for (int i = 0;  i < ledige.length;  ++i)
	    System.out.printf("%4d sengeplasser: %6d\n", i+1, ledige[i]);
    }
}

class RomTeller implements Runnable {
    Hotell hotell;
    Monitor mon;
    CountDownLatch teller;

    RomTeller (Hotell h, Monitor m, CountDownLatch c) {
	hotell = h;  mon = m;  teller = c;
    }
    
    @Override
    public void run () {
	mon.rapporterLedigeRom(hotell.ledigeRom());
	teller.countDown();
    }
}

class Monitor {
    Lock las = new ReentrantLock();
    int[] sumLedige;

    Monitor (int maxSeng) {
	sumLedige = new int[maxSeng];
    }

    void rapporterLedigeRom(int[] ledige) {
	las.lock();
	// Trenger ikke try-finally siden ingen trådoperasjoner i denne løkken.
	for (int i = 0;  i < ledige.length;  ++i)
	    sumLedige[i] += ledige[i];
	las.unlock();
    }

    int[] hentLedigeRom() {
	// Trenger ikke beskyttes siden alle produsenter er ferdige.
	return sumLedige;
    }
}

class IngenReservasjon extends RuntimeException {
    IngenReservasjon (String navn) {
	super(navn + " har ingen reservasjon");
    }
}

class IntetLedigRom extends RuntimeException {
    IntetLedigRom (String navn, int antSeng) {
	super("Intet ledig rom med " + antSeng + " sengeplasser til " + navn);
    }
}

class Gjest {
    String navn;
    Rom rom = null;
}

class Reservasjon {
    Gjest gjest;
    int trengerSenger;
    boolean onskerKjokken;
    Reservasjon nesteR, forrigeR;
}
